/*
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	Copyright (C) 2010 John Crispin <blogic@openwrt.org>
 */

#include <limits.h>

#include <pj/log.h>

#include "backend.h"
#include "../owsip.h"
#include "../config.h"
#include "../call/call.h"
#include "../dial/detector.h"
#include "../dial/dial.h"
#include "../notify/notify.h"

#define THIS_FILE	"b/ltq-tapi.c"

enum state
{
	ONHOOK = 0,	// on hook
	RING,		// on hook ; ringing
	OFFHOOK,	// off hook
	DIAL,		// off hook ; started dialing
	CONNECT,	// off hook ; call is connecting
	CALL,		// off hook ; in a call
	END,		// off hook ; call ended
	ERROR,		// error occured
};

struct ltq_tapi_port
{
	enum state state;	// the state of the port
	struct port *config;	// the ports uci config
	struct call *call;	// the call that we are handling, if any
	struct detector detector;	// we feed the detector digits and get numbers
	pj_timer_entry ring_timer;	// the timer used to generate the ring intervals
	pjsua_conf_port_id id;		// the id of the conference bridge
};

extern void (*tapi_digit_callback)(pj_uint8_t port, pj_uint8_t digit);
extern void (*tapi_hook_callback)(pj_uint8_t port, pj_uint8_t event);

extern pj_status_t tapi_hook_status(pj_uint8_t port, pj_int32_t *status);
extern pj_status_t tapi_ring(pj_uint8_t port, pj_uint8_t state, char *caller_number);
extern pj_status_t tapi_tone(pj_uint8_t port, pj_uint8_t code);
extern void tapi_revert_channels(void);
extern void tapi_cid_select(char *cid);
extern void tapi_locale_select(char *country);

#define MAX_PORT		2
#define TONE_BUSY		33
#define TONE_CONGESTION		34
#define TONE_DIAL		35
#define TONE_RING		36
#define TONE_WAITING		37

static struct ltq_tapi_port ports[MAX_PORT];
static struct call *ring[MAX_PORT];

enum {
	LED_OFF = 0,
	LED_ON,
	LED_BLINK
};

static void
led_set(int port, int value)
{
	FILE *fp;
	char path[NAME_MAX];

	if (!ports[port].config || !ports[port].config->led)
		return;

	snprintf(path, NAME_MAX, "/sys/class/leds/%s/trigger", ports[port].config->led);
	fp = fopen(path, "w");
	if (!fp) {
		pjsua_perror(__FILE__, "opening led path failed", PJ_ENOTFOUND);
		return;
	}
	if (value == LED_BLINK)
		fprintf(fp, "timer");
	else
		fprintf(fp, "none");
	fclose(fp);

	if (value != LED_BLINK) {
		snprintf(path, NAME_MAX, "/sys/class/leds/%s/brightness", ports[port].config->led);

		fp = fopen(path, "w");
		if (!fp) {
			pjsua_perror(__FILE__, "opening led path failed", PJ_ENOTFOUND);
			return;
		}
		fprintf(fp, "%d", value);
		fclose(fp);
	}
}

static void
ltq_tapi_port_state(int port, int state, struct call *c)
{
	char caller_number[15 + 1] = { 0 };
	int old_state;
	pjsua_call_info ci;

	old_state = ports[port].state;
	ports[port].state = state;


	switch (state) {
	case ONHOOK:
		PJ_LOG(3, (THIS_FILE, "port %d on hook", port));
		notify_port("on_hook", port);
		if(((old_state == CALL) || (old_state == CONNECT)) && ports[port].call) {
			owsip_hangup(ports[port].call->call_id);
			call_deassoc_port(ports[port].call->call_id);
		}
		tapi_tone(port, 0);
		tapi_ring(port, 0, NULL);
		ports[port].state = ONHOOK;
		ring[port] = NULL;
		ports[port].call = NULL;
		led_set(port, LED_OFF);
		break;

	case OFFHOOK:
		PJ_LOG(3, (THIS_FILE, "port %d off hook", port));
		notify_port("off_hook", port);
		ports[port].state = OFFHOOK;
		tapi_tone(port, TONE_DIAL);
		led_set(port, LED_ON);
		break;

	case RING:
		PJ_LOG(3, (THIS_FILE, "port %d ring", port));
		ring[port] = c;
		led_set(port, LED_BLINK);
		if (ports[port].config && ports[port].config->noring)
			break;
		pjsua_call_get_info(c->call_id, &ci);
		sscanf(ci.remote_info.ptr, "\"%15[^\"]s", caller_number);
		if (strncmp("Anonymous", caller_number, strlen("Anonymous")) != 0)
			tapi_ring(port, 1, caller_number);
		else
			tapi_ring(port, 1, NULL);
		break;

	case CONNECT:
		PJ_LOG(3, (THIS_FILE, "port %d connect", port));
		tapi_tone(port, 0);
		ports[port].call = c;
		call_assoc_port(c->call_id, port);
		led_set(port, LED_ON);
		break;

	case DIAL:
		PJ_LOG(3, (THIS_FILE, "port %d dial", port));
		tapi_tone(port, 0);
		break;

	case CALL:
		PJ_LOG(3, (THIS_FILE, "port %d call", port));
		tapi_tone(port, TONE_RING);
		led_set(port, LED_BLINK);
		break;

	case ERROR:
		PJ_LOG(3, (THIS_FILE, "port %d error", port));
		ports[port].call = NULL;
		tapi_tone(port, TONE_CONGESTION);
		break;

	case END:
		PJ_LOG(3, (THIS_FILE, "port %d end", port));
		ports[port].call = NULL;
		tapi_tone(port, TONE_CONGESTION);
		break;

	default:
		PJ_LOG(2, (THIS_FILE, "port %d entered unknown state %d", port, state));
		break;
	}
}

static void
ltq_tapi_do_ring(struct call *call, int port, int state)
{
	int i;
	for (i = 0; i < MAX_PORT; i++) {
		if ((port >= 0) && (port != i))
			continue;
		switch (state) {
		case 0:
			if ((ring[i] != call) || (ports[i].state != RING))
				break;
			ltq_tapi_port_state(i, ONHOOK, NULL);
			break;
		default:
			if (ring[i] || (ports[i].state != ONHOOK))
				break;
			ltq_tapi_port_state(i, RING, call);
			break;
		}
	}
}

static void
ltq_tapi_pickup(struct call *call, int port)
{
	ports[port].call = call;
	ltq_tapi_do_ring(call, -1, 0);
	ltq_tapi_port_state(port, CONNECT, call);
	owsip_pickup(call->call_id);
}

static void
ltq_tapi_detector_event(struct detector *d, enum detector_event e, char data)
{
	int port = (int)d->priv;

	switch(e) {
		case DETECTOR_HOOK:
			PJ_LOG(3, (THIS_FILE, "%s hook on port %d", (data)?("off"):("on"), port));
			if ((data) && (ports[port].state == RING) && ring[port])
				ltq_tapi_pickup(ring[port], port);
			else if (data)
				ltq_tapi_port_state(port, OFFHOOK, NULL);
			else
				ltq_tapi_port_state(port, ONHOOK, NULL);
			break;

		case DETECTOR_DIGIT:
			break;
	}
}

static void
ltq_tapi_detector_number(struct detector *d, char *nr)
{
	struct call *c;
	int port = (int)d->priv;
	c = dial_call_nr(port, nr);
	if (c)
		ltq_tapi_port_state(port, CONNECT, c);
	else
		ltq_tapi_port_state(port, ERROR, NULL);
}

void ltq_tapi_digit(pj_uint8_t port, pj_uint8_t digit)
{
	PJ_LOG(3, (THIS_FILE, "port %d detected tone %c", port, digit));
	if (ports[port].state == OFFHOOK)
		ltq_tapi_port_state(port, DIAL, NULL);
	detector_feed(&ports[port].detector, DETECTOR_DIGIT, digit);
}

void ltq_tapi_hook(pj_uint8_t port, pj_uint8_t event)
{
	detector_feed(&ports[port].detector, DETECTOR_HOOK, event);
}

static void
ltq_tapi_early_init(struct owsip *owsip)
{
	if (owsip->conf.revert)
		tapi_revert_channels();
	if (owsip->conf.cid)
		tapi_cid_select(owsip->conf.cid);
}

static void
ltq_tapi_init(struct owsip *owsip)
{
	pjmedia_aud_dev_info info[16];
	pj_uint32_t count = 16;
	pj_int32_t hook_status, i;

	tapi_digit_callback = ltq_tapi_digit;
	tapi_hook_callback = ltq_tapi_hook;

	pjsua_enum_aud_devs(info, &count);
	for(i = 0; i < count; i++)
		PJ_LOG(3, (THIS_FILE, "initializing audio device %d named %s", i, info[i].name));
	pjsua_set_snd_dev(0, 0);
	ports[0].id = 0;
	pjsua_add_snd_port(1, &ports[1].id);
	for(i = 0; i < MAX_PORT; i++) {
		struct list_head *p;
		list_for_each(p, &owsip->ports) {
			struct port *po = list_entry(p, struct port, list);
			if (po->id == i)
				ports[i].config = po;
		}
		ports[i].id = i;
		ports[i].detector.event = ltq_tapi_detector_event;
		ports[i].detector.number = ltq_tapi_detector_number;
		ports[i].detector.priv = (void *)i;
		detector_init(&ports[i].detector);
		tapi_hook_status(i, &hook_status);
		if (hook_status)
			ltq_tapi_port_state(i, OFFHOOK, NULL);
	}
}

static void
ltq_tapi_late_init(struct owsip *owsip)
{
	if (owsip->conf.locale)
		tapi_locale_select(owsip->conf.locale);
}

static void
ltq_tapi_connect(struct call *c)
{
	pjsua_call_info ci;
	pjsua_call_get_info(c->call_id, &ci);
	if (ci.media_status == PJSUA_CALL_MEDIA_ACTIVE) {
		PJ_LOG(3, (THIS_FILE, "connect call to port %d", c->port));
		pjsua_conf_connect(ci.conf_slot, ports[c->port].id);
		pjsua_conf_connect(ports[c->port].id, ci.conf_slot);
	}
}

static void
ltq_tapi_disconnect(struct call *c)
{
	pjsua_call_info ci;
	pjsua_call_get_info(c->call_id, &ci);
	if (ci.media_status == PJSUA_CALL_MEDIA_ACTIVE)
		pjsua_conf_disconnect(ci.conf_slot, ports[c->port].id);
}

static void
ltq_tapi_media_config(pjsua_media_config *media_cfg)
{
	media_cfg->clock_rate = 8000;
	media_cfg->snd_auto_close_time = -1;
}

static void
ltq_tapi_ring(struct call *c)
{
	OWSIP_TRACER("enter");

	ltq_tapi_do_ring(c, c->account->port, 1);
}

static void
ltq_tapi_call(struct call *c)
{
	OWSIP_TRACER("enter");

	if (c->port >= 0)
		ltq_tapi_port_state(c->port, CALL, c);
}

static void
ltq_tapi_stop(struct call *c)
{
	OWSIP_TRACER("enter");

	if ((c->port >= 0) && (ports[c->port].state != ONHOOK))
	        ltq_tapi_port_state(c->port, END, NULL);
        ltq_tapi_do_ring(c, -1, 0);

}

static struct backend ltq_tapi_backend = {
	.ports = MAX_PORT,
	.name = "ltq_tapi",
	.early_init = ltq_tapi_early_init,
	.late_init = ltq_tapi_late_init,
	.init = ltq_tapi_init,
	.connect = ltq_tapi_connect,
	.disconnect = ltq_tapi_disconnect,
	.media_config = ltq_tapi_media_config,
	.ring = ltq_tapi_ring,
	.call = ltq_tapi_call,
	.stop = ltq_tapi_stop,
};

BACKEND(ltq_tapi_backend)
