/*
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	Copyright (C) 2010 John Crispin <blogic@openwrt.org>
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#include "config.h"
#include "dial/dial.h"
#include "notify/notify.h"

struct dialuci {
	struct ucimap_section_data map;
	char *prefix;
	char *handler;
};

struct stun {
	struct ucimap_section_data map;
	char *host;
	int port;
};

static struct owsip owsip;

static struct account*
account_find(char *name)
{
	struct list_head *p;

        list_for_each(p, &owsip.accounts) {
		struct account *a = list_entry(p, struct account, list);
		if (!strcmp(name, a->name))
			return a;
	}
	return NULL;
}

static int
account_init(struct uci_map *map, void *section, struct uci_section *s)
{
	struct account *a = section;
	a->name = strdup(s->e.name);
	return 0;
}

static int
account_add(struct uci_map *map, void *section)
{
	struct account *a = section;
	if(!a->realm || !a->username || !a->password) {
		printf("settings for account '%s' are incomplete\n", a->name);
		exit(-1);
	}
	a->port = -1;
	if(a->cport)
		a->port = atoi(a->cport);
	if(!a->disabled)
		list_add_tail(&a->list, &owsip.accounts);
	return 0;
}

static struct uci_optmap account_uci_map[] =
{
	{
		UCIMAP_OPTION(struct account, realm),
		.type = UCIMAP_STRING,
		.name = "realm",
	},{
		UCIMAP_OPTION(struct account, username),
		.type = UCIMAP_STRING,
		.name = "username",
	},{
		UCIMAP_OPTION(struct account, password),
		.type = UCIMAP_STRING,
		.name = "password",
	},{
		UCIMAP_OPTION(struct account, cport),
		.type = UCIMAP_STRING,
		.name = "port",
	},{
		UCIMAP_OPTION(struct account, disabled),
		.type = UCIMAP_INT,
		.name = "disabled",
	},{
		UCIMAP_OPTION(struct account, def_acc),
		.type = UCIMAP_INT,
		.name = "default",
	},
};

static struct uci_sectionmap account_sectionmap = {
	UCIMAP_SECTION(struct account, map),
	.type = "account",
	.init = account_init,
	.add = account_add,
	.options = account_uci_map,
	.n_options = ARRAY_SIZE(account_uci_map),
	.options_size = sizeof(struct uci_optmap),
};

static int
conf_init(struct uci_map *map, void *section, struct uci_section *s)
{
	return 0;
}

static int
conf_add(struct uci_map *map, void *section)
{
	struct owsip_conf *l = section;
	if (!l->backend)
		l->backend = backend_get_default();
	if(!l->backend) {
		printf("no backend specified in config file\n");
		exit(-1);
	}
	memcpy(&owsip.conf, section, sizeof(struct owsip_conf));
	free(l);
	return 0;
}

static struct uci_optmap conf_uci_map[] =
{
	{
		UCIMAP_OPTION(struct owsip_conf, name),
		.type = UCIMAP_STRING,
		.name = "name",
	},{
		UCIMAP_OPTION(struct owsip_conf, backend),
		.type = UCIMAP_STRING,
		.name = "backend",
	},{
		UCIMAP_OPTION(struct owsip_conf, ossdev),
		.type = UCIMAP_INT,
		.name = "ossdev",
	},{
		UCIMAP_OPTION(struct owsip_conf, log_level),
		.type = UCIMAP_INT,
		.name = "log_level",
	},{
		UCIMAP_OPTION(struct owsip_conf, interface),
		.type = UCIMAP_STRING,
		.name = "interface",
	},{
		UCIMAP_OPTION(struct owsip_conf, local_port),
		.type = UCIMAP_INT,
		.name = "local_port",
	},{
		UCIMAP_OPTION(struct owsip_conf, rtp_port),
		.type = UCIMAP_INT,
		.name = "rtp_port",
	},{
		UCIMAP_OPTION(struct owsip_conf, outbound),
		.type = UCIMAP_STRING,
		.name = "outbound",
	},{
		UCIMAP_OPTION(struct owsip_conf, revert),
		.type = UCIMAP_INT,
		.name = "revert",
	},{
		UCIMAP_OPTION(struct owsip_conf, cid),
		.type = UCIMAP_STRING,
		.name = "cid",
	},{
		UCIMAP_OPTION(struct owsip_conf, locale),
		.type = UCIMAP_STRING,
		.name = "locale",
	},
};

static struct uci_sectionmap conf_sectionmap = {
	UCIMAP_SECTION(struct owsip_conf, map),
	.type = "general",
	.init = conf_init,
	.add = conf_add,
	.options = conf_uci_map,
	.n_options = ARRAY_SIZE(conf_uci_map),
	.options_size = sizeof(struct uci_optmap),
};

static int
stun_init(struct uci_map *map, void *section, struct uci_section *s)
{
	return 0;
}

static int
stun_add(struct uci_map *map, void *section)
{
	struct stun *s = section;
	if(s->host && s->port) {
		char uri[256];
		snprintf(uri, 256, "%s:%d", s->host, s->port);
		owsip.cfg.stun_srv[owsip.cfg.stun_srv_cnt++] = pj_str(uri);
	} else {
		printf("failed to load stun server %s:%d\n", s->host, s->port);
	}
	if(s->host)
		free(s->host);
	free(s);
	return 0;
}

static struct uci_optmap stun_uci_map[] =
{
	{
		UCIMAP_OPTION(struct stun, host),
		.type = UCIMAP_STRING,
		.name = "host",
	},{
		UCIMAP_OPTION(struct stun, port),
		.type = UCIMAP_INT,
		.name = "port",
	},
};

static struct uci_sectionmap stun_sectionmap = {
	UCIMAP_SECTION(struct stun, map),
	.type = "stun",
	.init = stun_init,
	.add = stun_add,
	.options = stun_uci_map,
	.n_options = ARRAY_SIZE(stun_uci_map),
	.options_size = sizeof(struct uci_optmap),
};

static int
dial_init(struct uci_map *map, void *section, struct uci_section *s)
{
	return 0;
}

static int
dial_add(struct uci_map *map, void *section)
{
	struct dialuci *d = section;
	if(d->handler && d->prefix)
		dial_set_prefix(d->prefix, d->handler);
	else
		printf("failed to load dial %s %s\n", d->prefix, d->handler);
	if(d->prefix)
		free(d->prefix);
	if(d->handler)
		free(d->handler);
	free(d);
	return 0;
}

static struct uci_optmap dial_uci_map[] =
{
	{
		UCIMAP_OPTION(struct dialuci, prefix),
		.type = UCIMAP_STRING,
		.name = "prefix",
	},{
		UCIMAP_OPTION(struct dialuci, handler),
		.type = UCIMAP_STRING,
		.name = "handler",
	},
};

static struct uci_sectionmap dial_sectionmap = {
	UCIMAP_SECTION(struct dialuci, map),
	.type = "dial",
	.init = dial_init,
	.add = dial_add,
	.options = dial_uci_map,
	.n_options = ARRAY_SIZE(dial_uci_map),
	.options_size = sizeof(struct uci_optmap),
};

static int
port_init(struct uci_map *map, void *section, struct uci_section *s)
{
	struct port *p = section;
	p->name = strdup(s->e.name);
	return 0;
}

static int
port_add(struct uci_map *map, void *section)
{
	struct port *p = section;
	if((p->id < 0) || (p->id > 1)) {
		printf("settings for port '%s' are incomplete\n", p->name);
		return -1;
	}
	list_add_tail(&p->list, &owsip.ports);
	return 0;
}

static struct uci_optmap port_uci_map[] =
{
	{
		UCIMAP_OPTION(struct port, id),
		.type = UCIMAP_INT,
		.name = "id",
	},{
		UCIMAP_OPTION(struct port, noring),
		.type = UCIMAP_INT,
		.name = "noring",
	},{
		UCIMAP_OPTION(struct port, nodial),
		.type = UCIMAP_INT,
		.name = "nodial",
	},{
		UCIMAP_OPTION(struct port, led),
		.type = UCIMAP_STRING,
		.name = "led",
	},
};

static struct uci_sectionmap port_sectionmap = {
	UCIMAP_SECTION(struct contact, map),
	.type = "port",
	.init = port_init,
	.add = port_add,
	.options = port_uci_map,
	.n_options = ARRAY_SIZE(port_uci_map),
	.options_size = sizeof(struct uci_optmap),
};

static int
contact_init(struct uci_map *map, void *section, struct uci_section *s)
{
	struct contact *c = section;
	c->name = strdup(s->e.name);
	return 0;
}

static void
contact_translate(char *n)
{
	while (*n) {
		if ((*n >= 'A') && (*n <= 'Z'))
			*n = '2' + ((*n - 'A') / 3);
		if ((*n >= 'a') && (*n <= 'z'))
			*n = '2' + ((*n - 'a') / 3);
		if (*n > '9')
			*n = '9';
		if (*n < '0')
			*n = '0';
		n++;
	}
}

static int
contact_add(struct uci_map *map, void *section)
{
	struct contact *c = section;
	if(!c->name || !c->code || !c->dial || !c->type) {
		printf("settings for contact '%s' are incomplete\n", c->name);
		return -1;
	}
	c->match = strdup(c->code);
	if (c->acc_name)
		c->account = account_find(c->acc_name);
	contact_translate(c->match);
	list_add_tail(&c->list, &owsip.contacts);
	return 0;
}

static struct uci_optmap contact_uci_map[] =
{
	{
		UCIMAP_OPTION(struct contact, desc),
		.type = UCIMAP_STRING,
		.name = "desc",
	},{
		UCIMAP_OPTION(struct contact, code),
		.type = UCIMAP_STRING,
		.name = "code",
	},{
		UCIMAP_OPTION(struct contact, dial),
		.type = UCIMAP_STRING,
		.name = "dial",
	},{
		UCIMAP_OPTION(struct contact, acc_name),
		.type = UCIMAP_STRING,
		.name = "account",
	},{
		UCIMAP_OPTION(struct contact, type),
		.type = UCIMAP_STRING,
		.name = "type",
	},
};

static struct uci_sectionmap contact_sectionmap = {
	UCIMAP_SECTION(struct contact, map),
	.type = "contact",
	.init = contact_init,
	.add = contact_add,
	.options = contact_uci_map,
	.n_options = ARRAY_SIZE(contact_uci_map),
	.options_size = sizeof(struct uci_optmap),
};

static int
rule_init(struct uci_map *map, void *section, struct uci_section *s)
{
	struct rule *r = section;
	r->name = strdup(s->e.name);
	return 0;
}

static int
rule_add(struct uci_map *map, void *section)
{
	struct rule *r = section;
	list_add_tail(&r->list, &owsip.rules);
	return 0;
}

static struct uci_optmap rule_uci_map[] =
{
	{
		UCIMAP_OPTION(struct rule, handler),
		.type = UCIMAP_STRING,
		.name = "handler",
	},{
		UCIMAP_OPTION(struct rule, account),
		.type = UCIMAP_STRING,
		.name = "account",
	},{
		UCIMAP_OPTION(struct rule, match),
		.type = UCIMAP_STRING,
		.name = "match",
	},{
		UCIMAP_OPTION(struct rule, start),
		.type = UCIMAP_INT,
		.name = "start",
	},{
		UCIMAP_OPTION(struct rule, stop),
		.type = UCIMAP_INT,
		.name = "stop",
	}
};

static struct uci_sectionmap rule_sectionmap = {
	UCIMAP_SECTION(struct rule, map),
	.type = "rule",
	.init = rule_init,
	.add = rule_add,
	.options = rule_uci_map,
	.n_options = ARRAY_SIZE(rule_uci_map),
	.options_size = sizeof(struct uci_optmap),
};

static int
policy_init(struct uci_map *map, void *section, struct uci_section *s)
{
	struct rule *r = section;
	r->name = strdup(s->e.name);
	return 0;
}

static int
policy_add(struct uci_map *map, void *section)
{
	struct rule *r = section;
	owsip.policy.name = strdup("policy");
	owsip.policy.handler = r->handler;
	owsip.policy.start = r->start;
	owsip.policy.stop = r->stop;
	free(r);
	return 0;
}

static struct uci_sectionmap policy_sectionmap = {
	UCIMAP_SECTION(struct rule, map),
	.type = "policy",
	.init = policy_init,
	.add = policy_add,
	.options = rule_uci_map,
	.n_options = ARRAY_SIZE(rule_uci_map),
	.options_size = sizeof(struct uci_optmap),
};

static struct uci_sectionmap *owsip_smap[] = {
	&conf_sectionmap,
	&port_sectionmap,
	&account_sectionmap,
	&stun_sectionmap,
	&contact_sectionmap,
	&dial_sectionmap,
	&policy_sectionmap,
	&rule_sectionmap,
};

static struct uci_map owsip_map = {
	.sections = owsip_smap,
	.n_sections = ARRAY_SIZE(owsip_smap),
};

struct owsip*
config_load(void)
{
	int ret;
	struct uci_context *ctx = uci_alloc_context();
	struct uci_package *pkg;

	INIT_LIST_HEAD(&owsip.ports);
	INIT_LIST_HEAD(&owsip.accounts);
	INIT_LIST_HEAD(&owsip.contacts);
	INIT_LIST_HEAD(&owsip.rules);

	ucimap_init(&owsip_map);
	ret = uci_load(ctx, "telephony", &pkg);
	if (ret)
		return NULL;
	ucimap_parse(&owsip_map, pkg);
	return &owsip;
}
