/*
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	Copyright (C) 2010 John Crispin <blogic@openwrt.org>
 */

#include <stdio.h>
#include <string.h>

#include <pj/log.h>

#include "dial.h"
#include "../call/call.h"

#define THIS_FILE       "d/dial.c"

static LIST_HEAD(diaplans);

#define MAX_PREFIX	12	// 0-9 * #
#define PREFIX_STAR	10
#define PREFIX_HASH	11
#define PREFIX_ERROR	255

static struct dial *prefixes[MAX_PREFIX];

void
dial_register(struct dial *b)
{
	if(!b || !b->name || !b->number) {
		printf("trying to register corrupt dial\n");
		return;
	}
	list_add(&b->list, &diaplans);
}

static unsigned char
dial_get_prefix(char *nr)
{
	unsigned char prefix = *nr - '0';
	if(*nr == '*') {
		prefix = PREFIX_STAR;
	} else if(*nr == '#') {
		prefix = PREFIX_HASH;
	} else if(prefix > PREFIX_HASH) {
		printf("invalid prefix %d\n", prefix);
		return PREFIX_ERROR;
	}
	return prefix;
}

static struct dial*
dial_find(char *name)
{
	struct list_head *p;
	list_for_each(p, &diaplans) {
		struct dial *d = container_of(p, struct dial, list);
		if(!strcmp(d->name, name))
			return d;
	}
	return 0;
}

void
dial_state(int state)
{
	struct list_head *p;
	list_for_each(p, &diaplans) {
		struct dial *d = container_of(p, struct dial, list);
		if((state == STATE_INIT) && d->init)
			d->init();
		else if((state == STATE_EXIT) && d->exit)
			d->exit();
	}
}

void
dial_set_prefix(char *nr, char *name)
{
	struct dial *d = dial_find(name);
	unsigned char prefix = dial_get_prefix(nr);
	if(prefix == PREFIX_ERROR) {
		printf("mapping invalid prefix %s to %s\n", nr, name);
		return;
	}
	if(!d) {
		printf("unknown dial prefix %s\n", name);
		return;
	}
	if(prefix >= MAX_PREFIX) {
		printf("out of range prefix %d\n", prefix);
		return;
	}
	if(prefixes[prefix]) {
		printf("prefix is already used by %s\n", prefixes[prefix]->name);
		return;
	}
	prefixes[prefix] = d;
}

struct call*
dial_call_nr(int port, char *nr)
{
	OWSIP_TRACER("enter");
	pjsua_call_id cid = -1;
	unsigned char prefix = dial_get_prefix(nr);
	struct call *c = NULL;

	PJ_LOG(3, (THIS_FILE, "port %d dialed %s", port, nr));

	if(prefix == PREFIX_ERROR) {
		PJ_LOG(2, (THIS_FILE, "invalid prefix %s was dialed", nr));
		return NULL;
	}
	if (prefix == PREFIX_STAR) {
		nr++;
		if (*nr) {
			struct dial *d = dial_find("contact");
			if (!d)
				return NULL;
			cid = d->number(nr);
			if (cid < 0 )
				return NULL;
		}
	} else if (prefix == PREFIX_HASH) {
		nr++;
		prefix = dial_get_prefix(nr);
		nr++;
		if(!prefixes[prefix]) {
			PJ_LOG(2, (THIS_FILE, "prefix %d is not mapped", prefix));
			return NULL;
		} else {
			cid = prefixes[prefix]->number(nr);
		}
	} else {
		if(!*nr) {
			PJ_LOG(2, (THIS_FILE, "number can not be NULL"));
			return NULL;
		}
		cid = owsip_call_realm(nr, owsip_account_map(port));
	}

	if (cid >= 0)
		c = call_assoc_port(cid, port);
	return c;
}
