/*
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	Copyright (C) 2010 John Crispin <blogic@openwrt.org>
 */

#include <fcntl.h>
#include <locale.h>
#include <signal.h>
#include <stdlib.h>
#include <unistd.h>

#include <arpa/inet.h>
#include <linux/netlink.h>
#include <linux/rtnetlink.h>
#include <net/if.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/types.h>

#include <libubox/list.h>
#include <libubox/uloop.h>

#include <pj/log.h>

#include "owsip.h"
#include "dial/dial.h"
#include "call/call.h"
#include "notify/notify.h"

static void netlink_new_msg(struct uloop_fd *ufd, unsigned events);

static struct owsip *owsip_ctx = 0;
static struct uloop_fd netlink_event = { .cb = netlink_new_msg };

struct contact*
owsip_contact(char *sequence)
{
	struct list_head *p;

	list_for_each(p, &owsip_ctx->contacts) {
		struct contact *c = list_entry(p, struct contact, list);
		if (!strcmp(sequence, c->match))
			return c;
	}
	return NULL;
}

struct account*
owsip_account_map(int port)
{
	struct list_head *p;

	list_for_each(p, &owsip_ctx->accounts) {
		struct account *a = list_entry(p, struct account, list);
		if (a->port == port)
			return a;
	}
	return list_first_entry(&owsip_ctx->accounts, struct account, list);
}

struct account*
owsip_account_id(pjsua_acc_id acc_id)
{
	struct list_head *p;
	list_for_each(p, &owsip_ctx->accounts) {
		struct account *a = list_entry(p, struct account, list);
		if (a->acc_id == acc_id)
			return a;
	}
	return NULL;
}

static void
on_incoming_call(pjsua_acc_id acc_id, pjsua_call_id call_id, pjsip_rx_data *rdata)
{
	pjsua_call_info ci;
	pjsua_call_get_info(call_id, &ci);
	struct account *acc = owsip_account_id(acc_id);
	PJ_LOG(3, (__FILE__, "incoming call from %.*s", (int)ci.remote_info.slen, ci.remote_info.ptr));
	if (acc) {
		pjsua_call_answer(call_id, 180, NULL, NULL);
		call_add(owsip_ctx, call_id, acc, CALL_RING);
	} else {
		PJ_LOG(2, (__FILE__, "failed to find account for id %d", acc_id));
	}
}

static void
on_call_state(pjsua_call_id call_id, pjsip_event *e)
{
	pjsua_call_info ci;
	pjsua_call_get_info(call_id, &ci);
	PJ_LOG(3, (__FILE__, "handling call with id %d and state %.*s", call_id, (int)ci.state_text.slen, ci.state_text.ptr));
	notify_call("call_state", call_id);
	switch(ci.state) {
	case PJSIP_INV_STATE_NULL:
		break;
	case PJSIP_INV_STATE_CALLING:
		call_add(owsip_ctx, call_id, owsip_account_id(ci.acc_id), CALL_CALL);
		break;
	case PJSIP_INV_STATE_INCOMING:
		break;
	case PJSIP_INV_STATE_EARLY:
		break;
	case PJSIP_INV_STATE_CONNECTING:
		break;
	case PJSIP_INV_STATE_CONFIRMED:
		break;
	case PJSIP_INV_STATE_DISCONNECTED:
		call_state(owsip_ctx, call_id, CALL_STOP);
		break;
	}
}

static void
on_call_media_state(pjsua_call_id call_id)
{
	pjsua_call_info ci;
	pjsua_call_get_info(call_id, &ci);
	call_state(owsip_ctx, call_id, CALL_CONNECT);
}

static void
on_nat_detect(const pj_stun_nat_detect_result *res)
{
	if(res->status != PJ_SUCCESS)
		pjsua_perror(__FILE__, "NAT detection failed", res->status);
	else
		PJ_LOG(3, (__FILE__, "NAT detected as %s", res->nat_type_name));
}

static pj_status_t
owsip_init(struct owsip *owsip)
{
	pj_status_t status;
	pjsua_media_config default_media_cfg;
	pjsua_codec_info codec[64];
	unsigned int codec_size = 64;
	int codec_count;

	owsip_ctx->backend = backend_get(owsip->conf.backend);
	if(!owsip_ctx->backend) {
		pjsua_perror(__FILE__, "failed to load backend", PJ_EINVAL);
		return PJ_EINVAL;
	} else {
		PJ_LOG(3, (__FILE__, "using backend %s", owsip_ctx->backend->name));
	}

	if(owsip_ctx->backend->early_init)
		owsip_ctx->backend->early_init(owsip_ctx);

	status = pjsua_create();
	if (status != PJ_SUCCESS) {
		pjsua_perror(__FILE__, "pjsua_create() failed", status);
		return status;
	}

	pjsua_config_default(&owsip->cfg);

	owsip->cfg.cb.on_incoming_call = &on_incoming_call;
	owsip->cfg.cb.on_call_media_state = &on_call_media_state;
	owsip->cfg.cb.on_call_state = &on_call_state;
	owsip->cfg.cb.on_nat_detect = &on_nat_detect;

	pjsua_logging_config_default(&owsip->log_cfg);
	owsip->log_cfg.console_level = owsip->conf.log_level;

	pjsua_media_config_default(&default_media_cfg);
	if(owsip_ctx->backend->media_config)
		owsip_ctx->backend->media_config(&default_media_cfg);

	if (!pjsua_verify_sip_url(owsip->conf.outbound)) {
		owsip->cfg.outbound_proxy[owsip->cfg.outbound_proxy_cnt++] = pj_str(owsip->conf.outbound);
	}

	status = pjsua_init(&owsip->cfg, (owsip->conf.log_level)?(&owsip->log_cfg):(NULL), &default_media_cfg);
	if (status != PJ_SUCCESS) {
		pjsua_perror(__FILE__, "pjsua_init() failed", status);
		return status;
	}

	pjsua_transport_config udp_cfg;
	pjsua_transport_config_default(&udp_cfg);

	if (owsip->conf.local_port) {
		udp_cfg.port = owsip->conf.local_port;
	} else {
		udp_cfg.port = 5060;

	}

	status = pjsua_transport_create(PJSIP_TRANSPORT_UDP, &udp_cfg, NULL);
	if (status != PJ_SUCCESS) {
		pjsua_perror(__FILE__, "pjsua_transport_create() failed", status);
		return status;
	}

	pjsua_transport_config rtp_cfg;
	pjsua_transport_config_default(&rtp_cfg);

	if (owsip->conf.rtp_port) {
		rtp_cfg.port = owsip->conf.rtp_port;
	} else {
		rtp_cfg.port = 4000;
	}

	status = pjsua_media_transports_create(&rtp_cfg);
	if (status != PJ_SUCCESS) {
		pjsua_perror(__FILE__, "pjsua_media_transports_create() failed", status);
		return status;
	}

	status = pjsua_call_subsys_start();
	if (status != PJ_SUCCESS)
		return status;

	status = pjsua_media_subsys_start();
	if (status != PJ_SUCCESS)
		return status;

	codec_count = pjsua_enum_codecs(codec, &codec_size);
	PJ_LOG(3, (__FILE__, "found %d codecs", codec_count));
	while(codec_count--) {
		PJ_LOG(3, (__FILE__, "codec %.*s has priority %d",
			(int)codec[codec_count].codec_id.slen,
			codec[codec_count].codec_id.ptr,
			codec[codec_count].priority));
	}

	if(owsip_ctx->backend->init)
		owsip_ctx->backend->init(owsip_ctx);

	if(owsip_ctx->backend->late_init)
		owsip_ctx->backend->late_init(owsip_ctx);

	dial_state(STATE_INIT);

	status = pjsua_pres_start();
	return status;
}

static pj_status_t
owsip_account(struct owsip *owsip)
{
	char id[512];
	char reg_uri[256];
	struct list_head *p;
	pj_status_t status;

	list_for_each(p, &owsip->accounts) {
		struct account *account = list_entry(p, struct account, list);

		pjsua_acc_config_default(&account->acc_config);

		snprintf(id, 512, "sip:%s@%s", account->username, account->realm);
		snprintf(reg_uri, 256, "sip:%s", account->realm);

		account->acc_config.id = pj_str(id);
		account->acc_config.reg_uri = pj_str(reg_uri);
		account->acc_config.cred_count = 1;
		account->acc_config.cred_info[0].realm = pj_str(account->realm);
		account->acc_config.cred_info[0].scheme = pj_str("digest");
		account->acc_config.cred_info[0].username = pj_str(account->username);
		account->acc_config.cred_info[0].data_type = PJSIP_CRED_DATA_PLAIN_PASSWD;
		account->acc_config.cred_info[0].data = pj_str(account->password);

		status = pjsua_acc_add(&account->acc_config, PJ_TRUE, &account->acc_id);
		if (status != PJ_SUCCESS)
			return status;

		if(account->def_acc)
			owsip->def_acc = account;
	}

	if (!owsip->def_acc) {
		owsip->def_acc = list_first_entry(&owsip->accounts, struct account, list);
	} else {
		status = pjsua_acc_set_default(owsip->def_acc->acc_id);
		if (status != PJ_SUCCESS)
			return status;
	}

	return PJ_SUCCESS;
}

pjsua_call_id
owsip_call(char *curi, struct account *a)
{
	pj_str_t uri;
	pj_status_t status = pjsua_verify_url(curi);
	pjsua_call_id call_id;

	if (status != PJ_SUCCESS) {
		PJ_LOG(2, (__FILE__, "can not call invalid URI: '%s'", curi));
		return -1;
	}
	if (!a)
		a = owsip_ctx->def_acc;
	uri = pj_str(curi);
	status = pjsua_call_make_call(a->acc_id, &uri, 0, NULL, NULL, &call_id);
	if (status != PJ_SUCCESS) {
		PJ_LOG(2, (__FILE__, "failed to call '%s'", curi));
		return -1;
	}
	return call_id;
}

pjsua_call_id
owsip_call_realm(char *curi, struct account *a)
{
	char buf[512] = { 0 };

	if (!a)
		a = owsip_ctx->def_acc;

	if (!a)
		return -1;

	snprintf(buf, 512, "sip:%s@%s", curi, a->realm);
	return owsip_call(buf, a);
}

pj_status_t
owsip_pickup(pjsua_call_id call_id)
{
	pjsua_call_info ci;
	pjsua_call_get_info(call_id, &ci);
	PJ_LOG(3, (__FILE__, "answering call from %.*s", (int)ci.remote_info.slen, ci.remote_info.ptr));
	return pjsua_call_answer(call_id, 200, NULL, NULL);
}

pj_status_t
owsip_hangup(pjsua_call_id call_id)
{
	pjsua_call_info ci;
	pjsua_call_get_info(call_id, &ci);
	PJ_LOG(3, (__FILE__, "hanging up call with %.*s", (int)ci.remote_info.slen, ci.remote_info.ptr));
	return pjsua_call_hangup(call_id, 410, NULL, NULL);
}

static void
owsip_netlink_interface(struct nlmsghdr *nlh)
{
	struct ifaddrmsg *ifa = (struct ifaddrmsg *) NLMSG_DATA(nlh);
	struct rtattr *rth = IFA_RTA(ifa);
	int rtl = IFA_PAYLOAD(nlh);

	while (rtl && RTA_OK(rth, rtl)) {
		if (rth->rta_type != IFA_LOCAL) {
			rth = RTA_NEXT(rth, rtl);
			continue;
		}

		uint32_t addr = htonl(*((uint32_t *)RTA_DATA(rth)));
		char if_name[IFNAMSIZ], if_addr[16];

		memset(&if_name, 0, sizeof(if_name));
		memset(&if_addr, 0, sizeof(if_addr));

		if_indextoname(ifa->ifa_index, if_name);

		if (strcmp(owsip_ctx->conf.interface, if_name)) {
			rth = RTA_NEXT(rth, rtl);
			continue;
		}

		snprintf(if_addr, 15, "%d.%d.%d.%d", (addr >> 24) & 0xff,
				(addr >> 16) & 0xff,
				(addr >> 8) & 0xff,
				addr & 0xff);

		PJ_LOG(3, (__FILE__, "interface %s has new ip %s", if_name, if_addr));

		pjsua_transport_config udp_cfg;
		pjsua_transport_config_default(&udp_cfg);

		udp_cfg.public_addr = pj_str(if_addr);
		udp_cfg.bound_addr = pj_str(if_addr);

		if ((pjsua_transport_create(PJSIP_TRANSPORT_UDP, &udp_cfg, NULL)) != PJ_SUCCESS)
			pjsua_perror(__FILE__, "pjsua_transport_create() failed", PJ_EINVAL);

		pjsua_transport_config rtp_cfg;
		pjsua_transport_config_default(&rtp_cfg);

		rtp_cfg.public_addr = pj_str(if_addr);
		rtp_cfg.bound_addr = pj_str(if_addr);

		if ((pjsua_media_transports_create(&rtp_cfg)) != PJ_SUCCESS)
			pjsua_perror(__FILE__, "pjsua_media_transports_create() failed", PJ_EINVAL);

		rth = RTA_NEXT(rth, rtl);
	}
}

static void
netlink_new_msg(struct uloop_fd *ufd, unsigned events)
{
	struct nlmsghdr *nlh;
	char buffer[BUFSIZ] = { 0 };
	int msg_size;

	nlh = (struct nlmsghdr *)buffer;
	if ((msg_size = recv(ufd->fd, nlh, BUFSIZ, 0)) == -1) {
		pjsua_perror(__FILE__, "error receiving netlink message", PJ_EINVAL);
		goto netlink_new_msg_done;
	}

	while (msg_size > sizeof(*nlh)) {
		int len = nlh->nlmsg_len;
		int req_len = len - sizeof(*nlh);

		if (req_len < 0 || len > msg_size) {
			pjsua_perror(__FILE__, "error reading netlink message", PJ_EINVAL);
			goto netlink_new_msg_done;
		}

		if (!NLMSG_OK(nlh, msg_size)) {
			pjsua_perror(__FILE__, "netlink message is not NLMSG_OK", PJ_EINVAL);
			goto netlink_new_msg_done;
		}

		if (nlh->nlmsg_type == RTM_NEWADDR)
			owsip_netlink_interface(nlh);

		msg_size -= NLMSG_ALIGN(len);
		nlh = (struct nlmsghdr*)((char*)nlh + NLMSG_ALIGN(len));
	}

netlink_new_msg_done:
	;
}

static pj_status_t
netlink_init(void)
{
	struct {
		struct nlmsghdr hdr;
		struct ifaddrmsg msg;
	} req;
	struct sockaddr_nl addr;
	int sock[2];

	memset(&addr, 0, sizeof(addr));
	memset(&req, 0, sizeof(req));

	if ((sock[0] = socket(PF_NETLINK, SOCK_RAW, NETLINK_ROUTE)) == -1) {
		pjsua_perror(__FILE__, "couldn't open NETLINK_ROUTE socket", PJ_EINVAL);
		return PJ_EINVAL;
	}

	addr.nl_family = AF_NETLINK;
	addr.nl_groups = RTMGRP_IPV4_IFADDR;
	if ((bind(sock[0], (struct sockaddr *)&addr, sizeof(addr))) == -1) {
		pjsua_perror(__FILE__, "couldn't bind netlink socket", PJ_EINVAL);
		return PJ_EINVAL;
	}

	netlink_event.fd = sock[0];
	uloop_fd_add(&netlink_event, ULOOP_READ | ULOOP_EDGE_TRIGGER);

	if ((sock[1] = socket(PF_NETLINK, SOCK_DGRAM, NETLINK_ROUTE)) == -1) {
		pjsua_perror(__FILE__, "couldn't open NETLINK_ROUTE socket", PJ_EINVAL);
		return PJ_EINVAL;
	}

	req.hdr.nlmsg_len = NLMSG_LENGTH(sizeof(struct ifaddrmsg));
	req.hdr.nlmsg_flags = NLM_F_REQUEST | NLM_F_ROOT;
	req.hdr.nlmsg_type = RTM_GETADDR;
	req.msg.ifa_family = AF_INET;

	if ((send(sock[1], &req, req.hdr.nlmsg_len, 0)) == -1) {
		pjsua_perror(__FILE__, "couldn't send netlink socket", PJ_EINVAL);
		return PJ_EINVAL;
	}

	struct uloop_fd dummy_event = { .fd = sock[1] };
	netlink_new_msg(&dummy_event, 0);

	return PJ_SUCCESS;
}

static void
signal_handler(int signal)
{
	PJ_LOG(1, (__FILE__, "exiting, received signal %d", signal));
	pjsua_destroy();
	uloop_done();
	exit(EXIT_FAILURE);
}

int
main(int argc, char *argv[])
{
	setlocale(LC_CTYPE, "");
	signal(SIGTERM, signal_handler);
	signal(SIGINT, signal_handler);
	signal(SIGPIPE, SIG_IGN);

	pj_status_t status;

	status = pj_init();
	if (status != PJ_SUCCESS)
		return EXIT_FAILURE;

	uci_state_dump();

	owsip_ctx = config_load();
	if (!owsip_ctx)
		return EXIT_FAILURE;

	status = owsip_init(owsip_ctx);
	if (status != PJ_SUCCESS)
		return EXIT_FAILURE;

	status = owsip_account(owsip_ctx);
	if (status != PJ_SUCCESS)
		return EXIT_FAILURE;

	uloop_init();

	status = netlink_init();
	if (status != PJ_SUCCESS)
		return EXIT_FAILURE;

	if(owsip_ctx->backend->mainloop)
		owsip_ctx->backend->mainloop(owsip_ctx);
	else
		uloop_run();

	pjsua_destroy();

	return EXIT_SUCCESS;
}
